"""MarkdownHelpFormatter."""

import argparse
import re

# -------------------------------------------------------------------------------
# /usr/lib/python3.8/argparse.py


class MarkdownHelpFormatter(argparse.HelpFormatter):
    """MarkdownHelpFormatter."""

    def __init__(self, *args, **kwargs):
        """Docstring."""
        kwargs["width"] = 89
        super().__init__(*args, **kwargs)
        self._is_new_section = None
        self._have_epilog = False
        self._debug = False

    # def _get_default_metavar_for_optional(self, action):
    #     return "=>" + action.dest.upper() + "<="

    # def _get_default_metavar_for_positional(self, action):
    #     return "==>" + action.dest + "<=="

    def add_argument(self, action):
        """Override."""
        # pylint: disable=protected-access
        if isinstance(action, argparse._SubParsersAction):
            if self._debug:
                print("ADD_ARGUMENT: action=_SubParsersAction", flush=True)
        else:
            if self._debug:
                print(f"ADD_ARGUMENT: action={action}", flush=True)
        return super().add_argument(action)

    def _add_item(self, func, args):
        if isinstance(args, tuple):
            action = args[1]  # line 265: args = usage, actions, groups, prefix
            # pylint: disable=protected-access
            if isinstance(action, argparse._SubParsersAction):
                if self._debug:
                    print("_ADD_ITEM: args.action=_SubParsersAction", flush=True)
            else:
                if self._debug:
                    print(f"_ADD_ITEM {action}", flush=True)
        super()._add_item(func, args)

    # -------------------------------------------------------------------------------

    #   def _metavar_formatter(self, action, default_metavar):
    #       if action.metavar is not None:
    #           result = action.metavar
    #       elif action.choices is not None:
    #           choice_strs = [str(choice) for choice in action.choices]
    #           # ylint: disable=consider-using-f-string
    #           result = "{%s}" % ",".join(choice_strs)
    #       else:
    #           result = default_metavar

    #       # result = f'<span style="color: darkred">***{result}***</span>'
    #       # result = f"***{result}***"

    #       # ylint: disable=redefined-builtin
    #       def format(tuple_size):
    #           # ylint: disable=no-else-return
    #           if isinstance(result, tuple):
    #               return result
    #           else:
    #               return (result,) * tuple_size

    #       return format

    def _format_usage(self, usage, actions, groups, prefix):
        if self._debug:
            print(f"_FORMAT_USAGE: usage={usage!r}", flush=True)
            print(f"_FORMAT_USAGE: type(self)={type(self)}", flush=True)
            print(f"_FORMAT_USAGE: actions={actions}", flush=True)
            print(f"_FORMAT_USAGE: groups={groups}", flush=True)
            print(f"_FORMAT_USAGE: prefix={prefix}", flush=True)
        string = super()._format_usage(usage, actions, groups, prefix)
        if self._debug:
            print(f"_FORMAT_USAGE: string={string!r}", flush=True)
        # Remove the 1st len("usage: ") chars from all lines.
        trim = len("usage: ")
        lines = [x[trim:] for x in string.rstrip().splitlines()]
        return "\n".join(
            [
                "### Usage",
                "",
                "```",
            ]
            + lines
            + [
                "```",
                "",
                "",
            ]
        )

    #   def format_help(self):
    #       """Docstring."""
    #       print(f"FORMAT_HELP: type(self)={type(self)}", flush=True)
    #       # print(self._prog)
    #       # self._root_section.heading = '# Options: %s' % self._prog
    #       return super().format_help()

    def start_section(self, heading):
        """Docstring."""

        # print(f"START SECTION {heading!r}")
        super().start_section(f"### {heading}")
        self._is_new_section = True

    # def end_section(self):
    #     """Docstring."""
    #     # print("END SECTION")
    #     super().end_section()

    def add_text(self, text):
        """Docstring."""

        if self._current_section == self._root_section:
            if self._is_new_section is not None and not self._have_epilog:
                self._have_epilog = True
                super().add_text("### Epilog")
            else:
                super().add_text("### Description")

        super().add_text(text)

    # -------------------------------------------------------------------------------

    def _format_action(self, action):
        # print(f"_FORMAT_ACTION: action.dest={action.dest}", flush=True)

        if action.dest == "command":
            return self._format_command_action(action)

        # print(f"action.default={action.default}", flush=True)
        if action.dest == "version":  # noqa: SIM114 Use logical or
            # version, new in 3.8, is suppressed by default.
            action.default = None

        elif action.dest == "help":
            # fix this - should not be necessary and incorrectly displays
            # what may have been intentionally hidden.
            action.default = None

        elif action.default == argparse.SUPPRESS:
            # print(f"suppressing {action.dest}", flush=True)
            return ""

        invocation = self._format_action_invocation(action)
        help_text = self._expand_help(action) if action.help else ""

        if self._is_new_section:
            self._is_new_section = False
            lines = [
                "| option | description |",
                "|:------ |:----------- |",
            ]
        else:
            lines = []

        # lines.append(f"| **{invocation}** | {help_text} |")
        lines.append(f"| `{invocation}` | {help_text} |")
        return "\n".join(lines) + "\n"

    # -------------------------------------------------------------------------------

    @staticmethod
    def _format_command_action(action):
        """New method."""

        lines = []

        for name, parser in action.choices.items():
            if (description := parser.description) is None:
                # print(f"Skipping={parser!r}")
                continue
            # print(f"\n=>name={name!r} parser={parser!r} description={description!r}")
            description = re.sub(" +", " ", description.replace("\n", " "))
            # print(parser.format_help())
            # print(f"\n=>name={name!r} parser={parser!r}")
            # print(f"| [{name}](docs/{name}.md) | {description} |")
            lines.append(f"| [`{name}`](docs/{name}.md) | {description} |")

        if len(lines) == 0:
            return ""

        # command = '<span style="color: darkred">***COMMAND***</span>'
        command = "COMMAND"
        return (
            "\n".join(
                [
                    f"| {command} | |",
                    "|:--- | --- |",
                ]
                + lines
            )
            + "\n\n"
        )



"""lscolors `chart` command."""

from lscolors.cmd import LscolorsCmd


class LscolorsChartCmd(LscolorsCmd):
    """lscolors `chart` command."""

    def init_command(self) -> None:
        """Initialize lscolors `chart` command."""

        parser = self.add_subcommand_parser(
            "chart",
            help="print color chart",
            description="Print color-chart.",
        )

        parser.add_argument(
            "color",
            nargs="*",
            help="print color palette with given colors",
        )

    def run(self) -> None:
        """Perform the command."""

        if self.options.color:
            self._print_palette()
            return

        self._chart_36x6_grey()
        # self._chart2()
        # self._chart_6_squares()
        # self._chart_216_lines()

    def _print_palette(self) -> None:
        for i in range(1, 4):
            for color in self.options.color:
                for _ in range(i):
                    text = "Lots of long sample text lorem ipsum dolar sit amet"
                    print(f"\033[{color}m{text} {color}\033[0m")

    @staticmethod
    def _chart_36x6_grey() -> None:
        offset = 16
        nlines, ncols = 36, 7
        maxcolor = 255  # nlines * ncols + offset - 1

        for i in range(nlines):
            line = ""
            for j in range(ncols):
                color = offset + (j * nlines) + i
                if color > maxcolor:
                    break
                ansi = f"38;05;{color:03}"
                if j > 0:
                    line += "    "
                line += f"\033[{ansi}m{ansi}\033[0m"
            print(line)

    @staticmethod
    def _chart2() -> None:
        for attr in range(8):
            line = ""
            for color in range(8):
                ansi = f"{attr:02};{color + 30:03}"
                if color > 0:
                    line += "    "
                line += f"\x1B[{ansi}m{ansi}\x1B[0m"
            print(line)

    @staticmethod
    def _chart_6_squares() -> None:
        def _6_squares() -> list[list[int]]:
            squares: list[list[int]] = [[], [], [], [], [], []]
            j = 0
            for i in range(216):
                if i > 0 and (i % 6) == 0:
                    j = (j + 1) % 6
                squares[j].append(16 + i)
            return squares

        for square in _6_squares():
            # print(square)
            for ansi in [f"38;05;{color:03}" for color in square]:
                print(f"\033[{ansi}m{ansi}\033[0m")

    @staticmethod
    def _chart_216_lines() -> None:
        # /usr/share/vim/vim81/syntax/dircolors.vim
        xterm_palette = ["00", "5f", "87", "af", "d7", "ff"]
        color = 16

        for red in xterm_palette:
            for green in xterm_palette:
                for blue in xterm_palette:
                    ansi = f"38;05;{color:03}"
                    print(f"\033[{ansi}m{color:03} = {ansi} #{red}{green}{blue}\033[0m")
                    color += 1  # noqa



"""lscolors `check` command."""

import lscolors.commands.utils.colors as colors_utils
import lscolors.commands.utils.config as config_utils
from lscolors.cmd import LscolorsCmd


class LscolorsCheckCmd(LscolorsCmd):
    """lscolors `check` command."""

    def init_command(self) -> None:
        """Initialize lscolors `check` command."""

        parser = self.add_subcommand_parser(
            "check",
            help="check database for required items",
            description="Check database in `$LS_COLORS` for required items.",
            epilog="Exit Status: zero indicates success, nonzero indicates failure.",
        )

        self.add_config_option(parser)
        self.add_colors_argument(parser)

    def run(self) -> None:
        """Perform the command."""

        config, meta_config = config_utils.load(self.options)

        try:
            colors, meta_colors = colors_utils.load(self.options)
        except RuntimeError as err:
            raise RuntimeError(
                f"{self.options.prog}: failure; {err}\n" f"{self.options.prog}: {meta_config}"
            ) from err

        required = (
            config["required_filenames"]
            + config["required_directories"]
            + config["required_extensions"]
        )
        nrequired = len(required)
        # missing = [x for x in required if x not in colors]
        missing = []
        for item in required:
            if item not in colors:
                missing.append(item)

        if (nmissing := len(missing)) > 0:
            raise RuntimeError(
                f"{self.options.prog}: failure; {len(colors)} items; "
                f"{nrequired - nmissing}/{nrequired} required; "
                f"{nmissing} missing {missing}\n"
                f"{self.options.prog}: {meta_colors}; {meta_config}"
            )

        print(
            f"{self.options.prog}: success; {len(colors)} items; "
            f"{nrequired}/{nrequired} required.\n"
            f"{self.options.prog}: {meta_colors}; {meta_config}"
        )



"""lscolors `samples` command."""

import os
import socket
import stat

from lscolors.cmd import LscolorsCmd
from lscolors.commands.utils import colors as colors_utils
from lscolors.commands.utils import config as config_utils
from lscolors.commands.utils import mkdir


class LscolorsSamplesCmd(LscolorsCmd):
    """lscolors `samples` command."""

    def init_command(self) -> None:
        """Initialize lscolors `samples` command."""

        parser = self.add_subcommand_parser(
            "samples",
            help="create directory of sample filesystem items",
            description=self.cli.dedent(
                """
                Create directory and populate with sample files, directories, etc.,
                for each item in `$LS_COLORS`, and all required items in
                configuration file `CONFIG`.
                """
            ),
        )

        self.add_config_option(parser)
        self.add_colors_argument(parser)

        arg = parser.add_argument(
            "--samplesdir",
            metavar="DIR",
            default="./lscolors-samples",
            help="create directory `DIR`",
        )
        self.cli.add_default_to_help(arg)

        parser.add_argument(
            "-f", "--force", action="store_true", help="Ok to clobber `DIR` if it exists"
        )

        #
        self._seqno = 0  # noqa

    def _fname(self, name: str) -> str:
        self._seqno += 1
        return f"{self._seqno:02}-{name}"

    def run(self) -> None:
        """Perform the command."""

        colors, meta_colors = colors_utils.load(self.options)
        config, meta_config = config_utils.load(self.options)

        mkdir.mkdir(self.options.samplesdir, self.options.force)
        print(
            self.options.prog,
            "creating directory",
            repr(self.options.samplesdir),
            "for",
            meta_colors,
            meta_config,
        )

        os.chdir(self.options.samplesdir)
        top = os.getcwd()

        # .../sample-files/colors/x.038-005-213.py
        path = os.path.join(top, "colors")
        os.mkdir(path)
        os.chdir(path)
        self._samples(colors, config, "x.{color}{name}")

        # .../sample-files/names/sample.py
        path = os.path.join(top, "names")
        os.mkdir(path)
        os.chdir(path)
        self._samples(colors, config, "sample{name}")

        # .../sample-files/hybrid/x.py.038-005-213.py
        path = os.path.join(top, "hybrid")
        os.mkdir(path)
        os.chdir(path)
        self._samples(colors, config, "x{name}.{color}{name}")

    def _samples(self, colors: dict[str, str], config: dict[str, list[str]], fmt: str) -> None:

        for name in config["required_filenames"]:
            if name[0] == "*":
                name = name[1:]
                with open(name, mode="w", encoding="utf-8"):
                    pass

        for name in config["required_directories"]:
            os.mkdir(name)

        for name in config["required_extensions"]:
            with open("required" + name, mode="w", encoding="utf-8"):
                pass

        self._create_basic_filetypes()
        self._create_extension_samples(colors, fmt)

        # sprinkle plenty of directory color around
        for alpha in "aiptz":
            os.mkdir(f"required.{alpha}-directory")
        for alpha in [".A", ".z", "A"]:
            os.mkdir(f"{alpha}-directory")

    def _create_basic_filetypes(self) -> None:
        """Create filesystem object for each "basic" filetype.

        Output from `dircolors --print-database`:

            #NORMAL 00 # no color code at all
            #FILE 00 # regular file: use no color at all
            RESET 0 # reset to "normal" color
            DIR 01;34 # directory
            LINK 01;36 # symbolic link. (If you set this to 'target' instead of a
            # numerical value, the color is as for the file pointed to.)
            MULTIHARDLINK 00 # regular file with more than one link
            FIFO 40;33 # pipe
            SOCK 01;35 # socket
            DOOR 01;35 # door
            BLK 40;33;01 # block device driver
            CHR 40;33;01 # character device driver
            ORPHAN 40;31;01 # symlink to nonexistent file, or non-stat'able file ...
            MISSING 00 # ... and the files they point to
            SETUID 37;41 # file that is setuid (u+s)
            SETGID 30;43 # file that is setgid (g+s)
            CAPABILITY 30;41 # file with capability
            STICKY_OTHER_WRITABLE 30;42 # dir that is sticky and other-writable (+t,o+w)
            OTHER_WRITABLE 34;42 # dir that is other-writable (o+w) and not sticky
            STICKY 37;44 # dir with the sticky bit set (+t) and not other-writable
            # This is for files with execute permission:
            EXEC 01;32

        Objects are named for `ls -l` to display in same order as template.
        """

        # pylint: disable=too-many-locals
        # pylint: disable=too-many-statements

        # NORMAL 00 # no color code at all
        normal_text = self._fname("NORMAL")
        with open(normal_text, mode="w", encoding="utf-8"):
            pass

        # FILE 00 # regular file: use no color at all
        regular_file = self._fname("FILE-regular-file")
        with open(regular_file, mode="w", encoding="utf-8"):
            pass

        # RESET 0 # reset to "normal" color
        reset_file = self._fname("RESET")
        with open(reset_file, mode="w", encoding="utf-8"):
            pass

        # DIR 00;36 # directory
        dir_not_ow = self._fname("DIR-not-other-writable")
        os.mkdir(dir_not_ow)

        # LINK 40;33 # symbolic link. (If you set this to 'target' instead of a
        # numerical value, the color is as for the file pointed to.)
        symlink_to_file = self._fname("LINK-symlink-file")
        os.symlink(regular_file, symlink_to_file)

        # MULTIHARDLINK 03 # regular file with more than one link
        hardlink_file1 = self._fname("MULTIHARDLINK-file1")
        with open(hardlink_file1, mode="w", encoding="utf-8"):
            pass

        # FIFO 03;31 # pipe
        fifo = self._fname("FIFO-named-pipe")
        os.mkfifo(fifo)

        # SOCK 03;32 # socket
        sockname = self._fname("SOCK-unix-domain-socket")
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(sockname)

        # DOOR 03;32 # door
        door = self._fname("DOOR-fake")
        with open(door, mode="w", encoding="utf-8"):
            pass

        # BLK 40;33;01 # block device driver
        blk_device = self._fname("BLK-block-device")
        try:
            os.mknod(blk_device, stat.S_IFBLK)
        except PermissionError:
            blk_device = blk_device + "-fake"
            with open(blk_device, mode="w", encoding="utf-8"):
                pass

        # CHR 40;33;01 # character device driver
        chr_device = self._fname("CHR-character-device")
        try:
            os.mknod(chr_device, stat.S_IFCHR)
        except PermissionError:
            chr_device = blk_device + "-fake"
            with open(chr_device, mode="w", encoding="utf-8"):
                pass

        # ORPHAN 40;31;03 # symlink to nonexistent file, or non-stat'able file
        symlink_to_orphan = self._fname("ORPHAN")
        os.symlink("nonexistent-file1", symlink_to_orphan)

        # MISSING 03 # ... and the files they point to
        # from `man 5 dir_colors`:
        #       (a nonexistent file which nevertheless
        #        has a symbolic link pointing to it)
        # I don't know what this means or how to create it.
        symlink_to_missing = self._fname("MISSING-fake")
        os.symlink("nonexistent-file2", symlink_to_missing)

        # SETUID 37;41 # file that is setuid (u+s)
        suid_file = self._fname("SETUID-file")
        with open(suid_file, mode="w", encoding="utf-8"):
            pass
        os.chmod(suid_file, stat.S_ISUID | 0o0755)

        # SETGID 30;43 # file that is setgid (g+s)
        sgid_file = self._fname("SETGID-file")
        with open(sgid_file, mode="w", encoding="utf-8"):
            pass
        os.chmod(sgid_file, stat.S_ISGID | 0o0755)

        # CAPABILITY 30;41 # file with capability
        cap_file = self._fname("CAPABILITY-file")
        with open(cap_file, mode="w", encoding="utf-8"):
            pass

        # STICKY_OTHER_WRITABLE 30;42 # dir that is sticky and other-writable (+t,o+w)
        sticky_dir_ow = self._fname("STICKY-dir-other-writable")
        os.mkdir(sticky_dir_ow)
        os.chmod(sticky_dir_ow, stat.S_ISVTX | 0o0777)

        # OTHER_WRITABLE 03;35 # dir that is other-writable (o+w) and not sticky
        dir_ow = self._fname("DIR-other-writable")
        os.mkdir(dir_ow)
        os.chmod(dir_ow, 0o0777)

        # STICKY 37;44 # dir with the sticky bit set (+t) and not other-writable
        sticky_dir = self._fname("STICKY-dir-not-other-writable")
        os.mkdir(sticky_dir)
        os.chmod(sticky_dir, stat.S_ISVTX | 0o0755)

        # EXEC 00 # This is for files with execute permission:
        exec_file = self._fname("EXEC-executable")
        with open(exec_file, mode="w", encoding="utf-8"):
            pass
        os.chmod(exec_file, 0o0755)

        # -- end of template --

        # complete MULTIHARDLINK
        os.link(hardlink_file1, self._fname("MULTIHARDLINK-file2"))

        # symlinks for everyone! set "LINK target" to see...
        os.symlink(symlink_to_file, self._fname("LINK-symlink-to-symlink-to-file"))
        os.symlink(fifo, self._fname("LINK-symlink-to-fifo"))
        os.symlink(sockname, self._fname("LINK-symlink-to-socket"))
        os.symlink(door, self._fname("LINK-symlink-to-door"))
        os.symlink(blk_device, self._fname("LINK-symlink-to-blk_device"))
        os.symlink(chr_device, self._fname("LINK-symlink-to-chr_device"))
        os.symlink(symlink_to_orphan, self._fname("LINK-symlink-to-symlink_to_orphan"))
        os.symlink(symlink_to_missing, self._fname("LINK-symlink-to-symlink_to_missing"))
        os.symlink(suid_file, self._fname("LINK-symlink-to-suid_file"))
        os.symlink(sgid_file, self._fname("LINK-symlink-to-sgid_file"))
        os.symlink(cap_file, self._fname("LINK-symlink-to-cap_file"))
        os.symlink(sticky_dir_ow, self._fname("LINK-symlink-to-sticky_dir_ow"))
        os.symlink(dir_ow, self._fname("LINK-symlink-to-dir_ow"))
        os.symlink(sticky_dir, self._fname("LINK-symlink-to-sticky_dir"))
        os.symlink(exec_file, self._fname("LINK-symlink-to-exec_file"))

    @staticmethod
    def _create_extension_samples(colors: dict[str, str], fmt: str) -> None:

        for name in [
            x
            for x in colors
            if x
            not in [
                "bd",
                "ca",
                "cd",
                "di",
                "do",
                "ex",
                "ln",
                "mh",
                "mi",
                "or",
                "ow",
                "pi",
                "rs",
                "sg",
                "so",
                "st",
                "su",
                "tw",
            ]
        ]:
            color = "-".join([f"{int(x):03}" for x in colors[name].split(";")])
            filename = fmt.format(color=color, name=name)
            try:
                os.stat(filename)
            except FileNotFoundError:
                with open(filename, mode="w", encoding="utf-8"):
                    pass



"""Lscolors base command."""

import argparse

from libcli import BaseCmd

import lscolors.commands.utils.colors
import lscolors.commands.utils.config


class LscolorsCmd(BaseCmd):
    """Lscolors base command class."""

    def init_command(self) -> None:
        """Initialize."""

    def run(self) -> None:
        """Perform."""

    def add_config_option(self, parser: argparse.ArgumentParser) -> None:
        """Add `--config` option."""
        lscolors.commands.utils.config.add_config_option(self.cli, parser)

    def add_colors_argument(self, parser: argparse.ArgumentParser) -> None:
        """Add `DIR_COLORS` argument."""
        lscolors.commands.utils.colors.add_colors_argument(self.cli, parser)



"""Create documentation for Subcommands."""

import argparse
import pathlib
import textwrap

from lscolors.cmd import LscolorsCmd
from lscolors.commands.utils import mkdir


class LscolorsDocsCmd(LscolorsCmd):
    """Create documentation for Subcommands."""

    def init_command(self) -> None:
        """Initialize create documentation for Subcommands."""

        parser = self.add_subcommand_parser(
            "docs",
            help="create documentation",
            description="Create documentation files for this application.",
            epilog="This is an internal command used during the packaging process.",
        )

        parser.set_defaults(
            format="markdown",
            docs="./docs",
        )

        parser.add_argument(
            "format",
            # nargs="?",
            choices=["ansi", "md", "txt"],
            help="Output format",
        )

        parser.add_argument(
            "docs",
            metavar="DIR",
            help="create directory `DIR`. " f"(default: {parser.get_default('docs')!r})",
        )

        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Ok to clobber `DIR` if it exists",
        )

    def run(self) -> None:
        """Perform the command."""

        mkdir.mkdir(self.options.docs, self.options.force)
        self.print_main_page(self.cli.parser)
        self.write_command_pages(self.cli.parser, self.options.docs)

    @staticmethod
    def print_main_page(main_parser: argparse.ArgumentParser) -> None:
        """Print main help page to `stdout`."""

        print(main_parser.format_help())

    def write_command_pages(self, main_parser: argparse.ArgumentParser, directory: str) -> None:
        """Create separate help pages for each command in `directory`."""

        # pylint: disable=protected-access
        assert main_parser._subparsers
        for action in main_parser._subparsers._actions:
            if isinstance(action, argparse._SubParsersAction):
                for name, parser in action.choices.items():
                    helptext = parser.format_help() + self._see_also()
                    pathlib.Path(directory, name + ".txt").write_text(helptext, encoding="utf-8")

    def _see_also(self) -> str:
        """Return string with refs to other pages."""

        # pylint: disable=protected-access
        parser = self.cli.parser
        also = {}
        assert parser._subparsers
        for action in parser._subparsers._actions:
            if isinstance(action, argparse._SubParsersAction):
                for name in action.choices:
                    also[name] = f"{parser.prog}-{name}"
        if not also:
            return ""

        formatter = parser._get_formatter()
        return "\n\nSee Also: \n" + textwrap.fill(
            ", ".join(also.values()) + ".",
            width=formatter._width,
            initial_indent=" " * formatter._indent_increment,
            subsequent_indent=" " * formatter._indent_increment,
        )



"""Command line interface."""

import sys

from libcli import BaseCLI

__all__ = ["LscolorsCLI"]


class LscolorsCLI(BaseCLI):
    """Command line interface."""

    def init_parser(self) -> None:
        """Initialize argument parser."""

        self.ArgumentParser(
            prog=__package__,
            description="Utilities for `dircolors(1)` and `dir_colors(5)`.",
            epilog="See `%(prog)s COMMAND --help` for help on a specific command.",
        )

    def add_arguments(self) -> None:
        """Add arguments to parser."""

        self.add_subcommand_modules("lscolors.commands", prefix="Lscolors", suffix="Cmd")

    def main(self) -> None:
        """Command line interface entry point (method)."""

        if not self.options.cmd:
            self.parser.print_help()
            self.parser.exit(2, "error: Missing COMMAND\n")

        try:
            self.options.cmd()
        except RuntimeError as err:
            print(err, file=sys.stderr)
            sys.exit(1)


def main(args: list[str] | None = None) -> None:
    """Command line interface entry point (function)."""
    LscolorsCLI(args).main()



"""lscolors database."""

import os
import subprocess
import sys
from argparse import ArgumentParser, Namespace

from libcli import BaseCLI


def add_colors_argument(cli: BaseCLI, parser: ArgumentParser) -> None:
    """Add arguments to parser."""

    _ = cli  # unused

    parser.add_argument(
        "dir_colors",
        nargs="?",
        metavar="DIR_COLORS",
        help="read file `DIR_COLORS` instead of `$LS_COLORS`",
    )


def load(options: Namespace) -> tuple[dict[str, str], str]:
    """Load color database from `options.dir_colors`, if given, or `$LS_COLORS`.

    Return multiple values:
        colors: dict, colors_by_filetype, k=filetype, v=color
        meta: str, identifies database loaded and, if relevant, the `$TERM` used.
    """

    meta = f"dir_colors={options.dir_colors!r}" if options.dir_colors else "env=$LS_COLORS"
    meta_with_term = meta + "; TERM=" + os.environ.get("TERM", "")

    if options.dir_colors:
        ls_colors = _compile_dir_colors(options, meta_with_term)
    elif not (_ls_colors := os.environ.get("LS_COLORS")):
        raise RuntimeError(f"missing `$LS_COLORS` environment variable; {meta}")
    else:
        ls_colors = _ls_colors

    if ls_colors:
        ls_colors = ls_colors.strip(":")
    if not ls_colors:
        raise RuntimeError(f"empty database; {meta_with_term}")

    #
    colors = {}
    for item in ls_colors.split(":"):
        filetype, color = item.split("=")
        if filetype.startswith("*."):
            filetype = filetype[1:]
        colors[filetype] = color

    return colors, meta


def _compile_dir_colors(options: Namespace, meta: str) -> str:
    """Run system `dircolors` (/usr/bin/dircolors) and return value of `LS_COLORS`."""

    proc = subprocess.run(
        ["dircolors", "--bourne-shell", options.dir_colors],
        capture_output=True,
        check=False,
        shell=False,
        text=True,
    )

    if proc.returncode:
        print(proc.stderr.strip(), file=sys.stderr)
        raise RuntimeError(f"system dircolors; {meta}")

    ls_colors = proc.stdout.strip()
    len1 = len("LS_COLORS='")
    len2 = len("';\nexport LS_COLORS")
    if len(ls_colors) >= len1 + len2:
        ls_colors = ls_colors[len1:-len2]

    return ls_colors



"""lscolors `paint` command."""

from __future__ import annotations

import re
import sys
from pathlib import Path
from typing import Iterable

from colors.colors import parse_rgb  # type: ignore

from lscolors.cmd import LscolorsCmd


class ColorGroup:
    """Color for collection of files."""

    _group_by_name: dict[str, ColorGroup] = {}
    _group_by_comment: dict[str, ColorGroup] = {}

    @classmethod
    def get_by_name(
        cls,
        name: str,
        default: ColorGroup | None = None,
    ) -> ColorGroup | None:
        """Return ColorGroup with matching `name`."""
        return cls._group_by_name.get(name, default)

    @classmethod
    def get_by_comment(
        cls,
        comment: str,
        default: ColorGroup | None = None,
    ) -> ColorGroup | None:
        """Return ColorGroup with matching `comment`."""
        return cls._group_by_comment.get(comment, default)

    @classmethod
    def items(cls) -> Iterable[ColorGroup]:
        """Return list of ColorGroup's in insertion order."""
        yield from cls._group_by_name.values()

    # https://stackoverflow.com/questions/11765623/convert-hex-to-closest-x11-color-number
    _closest_colors = []
    for i, num in enumerate([47, 68, 40, 40, 40, 21]):
        _closest_colors.extend([i] * num)
    _re_rgb = re.compile(r"[0-9A-Fa-f]{6}")

    @classmethod
    def rgb_to_xterm(cls, rgb: tuple[int, int, int]) -> int:
        """Return xterm color index from rgb tuple."""

        red = rgb[0]
        green = rgb[1]
        blue = rgb[2]
        max_ = max(red, green, blue)
        min_ = min(red, green, blue)

        if (max_ - min_) * (max_ + min_) <= 6250:
            color = 24 - (252 - ((red + green + blue) // 3)) // 10
            if 0 <= color <= 23:
                return 232 + color

        return (
            16
            + 36 * cls._closest_colors[red]
            + 6 * cls._closest_colors[green]
            + cls._closest_colors[blue]
        )

    def __init__(self, name: str, comment: str | None = None) -> None:
        """Add new ColorGroup to pool.

        Args:
            name: name of ColorGroup.
            comment: line to match in `dir_colors(5)` file.
        """

        self.group_name = name
        self.comment = comment
        self.color_name = None
        self.ansi: str | None = None
        self.rgb: tuple[int, int, int] | None = None

        self.__class__._group_by_name[self.group_name] = self
        if self.comment:
            self.__class__._group_by_comment[self.comment] = self

    def set_color(self, color: tuple[int, int, int]) -> None:
        """Assign `color` to this ColorGroup."""

        try:
            self.rgb = color
            rgb = parse_rgb(self.rgb)
            self.ansi = "38;5;" + str(ColorGroup.rgb_to_xterm(rgb))
        except ValueError:
            self.ansi = str(color)


class LscolorsPaintCmd(LscolorsCmd):
    """lscolors `paint` command."""

    # default groups
    ColorGroup("archive", " # archives or compressed (bright red)")
    ColorGroup("image", "# image formats")
    ColorGroup("audio", "# audio formats")

    # our additions
    ColorGroup("data", "# data")
    ColorGroup("source", "# source")
    ColorGroup("config", "# config")
    ColorGroup("history", "# history")
    ColorGroup("doc", "# doc")

    for filetype in [
        "NORMAL",
        "FILE",
        "RESET",
        "DIR",
        "LINK",
        "MULTIHARDLINK",
        "FIFO",
        "SOCK",
        "DOOR",
        "BLK",
        "CHR",
        "ORPHAN",
        "MISSING",
        "SETUID",
        "SETGID",
        "CAPABILITY",
        "STICKY_OTHER_WRITABLE",
        "OTHER_WRITABLE",
        "STICKY",
        "EXEC",
    ]:
        ColorGroup(filetype)

    def init_command(self) -> None:
        """Initialize lscolors `paint` command."""

        parser = self.add_subcommand_parser(
            "paint",
            help="paint dircolors",
            description="Apply palette to dircolors.",
        )

        arg = parser.add_argument(
            "dir_colors",
            nargs="?",
            metavar="DIR_COLORS",
            type=Path,
            default=Path.home() / ".dircolors",
            help="read `DIR_COLORS` file",
        )
        self.cli.add_default_to_help(arg)

        arg = parser.add_argument(
            "--encoding",
            metavar="NAME",
            default="utf-8",
            help="file encoding",
        )
        self.cli.add_default_to_help(arg)

        arg = parser.add_argument(
            "--palettes-dir",
            metavar="DIR",
            type=Path,
            default=Path.home() / ".palettes",
            help="directory of `coloors.co` palettes",
        )
        self.cli.add_default_to_help(arg)

        parser.add_argument(
            "--palette-num",
            metavar="NUMBER",
            type=int,
            help="id of `coloors.co` palette file to apply",
        )

        parser.add_argument(
            "--palette-file",
            metavar="FILE",
            type=Path,
            help="`coloors.co` palette file to apply",
        )

        parser.add_argument(
            "--pick",
            metavar="COLORNUM",
            type=int,
            nargs="+",
            help="select and order colors by palette-`COLORNUM`",
        )

        arg = parser.add_argument(
            "--add-samples",
            action="store_true",
            default=False,
            help="add color samples",
        )
        self.cli.add_default_to_help(arg)

        group_names = ", ".join([f"`{x.group_name}`" for x in ColorGroup.items()])
        parser.add_argument(
            "--group-color",
            metavar="GROUP=COLOR",
            nargs="+",
            help=f"paint `GROUP` with `COLOR`, where `GROUP` is one of {group_names}",
        )

    # -------------------------------------------------------------------------------

    def run(self) -> None:
        """Perform the command."""

        if self.options.palette_file:
            self._load_palette()

        for group_color in self.options.group_color or []:
            name, color = group_color.split("=")
            if (group := ColorGroup.get_by_name(name)) is not None:
                group.set_color(color)

        for i, color_group in enumerate(ColorGroup.items(), start=1):
            if not color_group.ansi:
                continue
            line = " ".join(
                [
                    f".COLOR-{i}",
                    f"{color_group.ansi:16}",
                    "#" + (str(color_group.rgb) or ""),
                    color_group.group_name,
                    color_group.comment or "",
                ]
            )
            print(f"\033[{color_group.ansi}m{line}\033[0m", file=sys.stderr)
            if self.options.add_samples:
                print(line)

        group_color = None
        # print("#", self.options.dir_colors)
        for line in self.options.dir_colors.read_text(self.options.encoding).splitlines():

            rstripped = line.rstrip()
            group_color = ColorGroup.get_by_comment(rstripped, group_color)

            stripped = rstripped.lstrip()
            if not stripped or stripped[0] == "#":
                print(rstripped)
                continue

            words = stripped.split(maxsplit=2)
            keyword = words[0]
            comment = words[2] if len(words) == 3 else None
            color = ColorGroup.get_by_name(keyword, group_color)

            if not color or not color.ansi:
                print(rstripped)
                continue

            parts = [keyword, color.ansi]
            if color.rgb:
                parts.append("#" + color.rgb)
            if comment:
                parts.append(comment)
            print(" ".join(parts))

    # -------------------------------------------------------------------------------

    def _load_palette(self) -> None:

        rgb_colors = self._read_palette()
        _len_rgb_colors = len(list(rgb_colors))
        _len_color_groups = len(list(ColorGroup.items()))

        if _len_rgb_colors < _len_color_groups:
            raise RuntimeError(
                f"Need {_len_color_groups} colors; "
                f"palette {str(self.options.palette_file)!r} has only {_len_rgb_colors}."
            )

        if self.options.pick:
            rgb_colors = [list(rgb_colors)[i - 1] for i in self.options.pick]

        for color_group, rgb in zip(ColorGroup.items(), rgb_colors):
            color_group.set_color(rgb)

    # -------------------------------------------------------------------------------

    def _read_palette(self) -> Iterable[tuple[int, int, int]]:
        """Parse `coolors.co` palette file and return list of colors.

        It's all in the first line; e.g. 3 colors
            `/* Coolors Exported Palette - https://coolors.co/00111c-001523-001a2c */`
        """

        if self.options.palette_num is not None:
            self.options.palette_file = (
                self.options.palettes_dir / f"palette ({self.options.palette_num}).txt"
            )

        palette = self.options.palette_file.read_text(self.options.encoding)
        lines = palette.splitlines()
        line = lines[0]

        prefix = "/* Coolors Exported Palette - https://coolors.co/"
        if not line.startswith(prefix):
            raise RuntimeError(f"Invalid prefix in palette `{str(self.options.palette_file)}`")

        suffix = " */"
        if not line.endswith(suffix):
            raise RuntimeError(f"Invalid suffix in palette `{str(self.options.palette_file)}`")

        clrs = [x.upper() for x in line[len(prefix) : -len(suffix)].split("-")]
        print("#", line[3:-2])
        return clrs



"""lscolors `sort` command module."""

import sys
from collections import defaultdict

from lscolors.cmd import LscolorsCmd


class LscolorsSortCmd(LscolorsCmd):
    """lscolors `sort` command class."""

    def init_command(self) -> None:
        """Initialize lscolors `sort` command."""

        self.add_subcommand_parser(
            "sort",
            help="sort lines of database file by color",
            description=self.cli.dedent(
                """
                Filter `stdin` to `stdout` sorting lines of a `DIR_COLORS` file
                by color then filetype. Blank lines and comments are unsorted
                and moved to the end.

                """
            ),
        )

    def run(self) -> None:
        """Perform the command."""

        _ = self  # unused

        # parse lines per https://github.com/coreutils/coreutils/blob/master/src/dircolors.c
        # which strips leading spaces, spaces between keyword and arg,
        # and trailing spaces after arg, which may contain spaces.

        lines_by_color = defaultdict(list)

        # save comments and empty lines for end; for modeline.
        comments = []

        for line in sys.stdin:
            line = line.strip()
            if not line or line[0] == "#":
                comments.append(line)
                continue

            arg = line.split(maxsplit=1)[1]  # (keyword, arg)

            if (i := arg.find("#")) >= 0:
                arg = arg[:i].strip()

            colors = []
            for color in arg.split(";"):
                colors.append(f"{int(color):03}")
            color = "-".join(colors)

            lines_by_color[color].append(line)

        for color in sorted(lines_by_color):
            for line in sorted(lines_by_color[color]):
                print(line)

        for line in comments:
            print(line)



"""lscolors `configs` command."""

import os.path

from lscolors.cmd import LscolorsCmd


class LscolorsConfigsCmd(LscolorsCmd):
    """lscolors `configs` command."""

    def init_command(self) -> None:
        """Initialize lscolors `configs` command."""

        self.add_subcommand_parser(
            "configs",
            help="print path to sample configuration files",
            description=str(
                "Print path to sample `.lscolors.yml` and `.dircolors` configuration files."
            ),
        )

    def run(self) -> None:
        """Perform the command."""

        _ = self  # unused

        print(os.path.join(os.path.dirname(os.path.dirname(__file__)), "sample-configs"))



"""Argparse help formatting."""

import argparse
import os
import pathlib

import icecream
import mandown.mandown
from icecream import ic

from . import ansi
from . import markdown

icecream.install()
ic.configureOutput(prefix="\nic ===> ", includeContext=True)


class Argformat:
    """Argparse help formatting."""

    # See /usr/lib/python3.8/argparse.py.

    formatter_class = ansi.AnsiHelpFormatter
    formatter_class_ext = ".ansi"

    def __init__(self, parser):
        """Initialize argparse help formatting.

        >>> import argparse
        >>> import argformat
        >>> parser = argparse.ArgumentParser(prog="hello")
        >>> argformat.argformat(parser)
        """

        if "_NO_ARGFORMAT" in os.environ:
            return

        parser.formatter_class = self.formatter_class
        parser._check_value = ansi.AnsiArgumentParser._check_value  # monkey patch

    # pylint: disable=redefined-builtin
    @classmethod
    def configure(cls, format):
        """Configure argparse help formatting."""

        if format == "ansi":
            cls.formatter_class = ansi.AnsiHelpFormatter
            cls.formatter_class_ext = ".ansi"
        elif format == "md":
            cls.formatter_class = markdown.MarkdownHelpFormatter
            cls.formatter_class_ext = ".md"
        else:
            cls.formatter_class = argparse.HelpFormatter
            cls.formatter_class_ext = ".txt"

    @classmethod
    def print_main_page(cls, main_parser):
        """Print main help page to `stdout`."""

        main_parser.formatter_class = cls.formatter_class
        print(main_parser.format_help())

    @classmethod
    def write_command_pages(cls, main_parser, directory):
        """Create separate help pages for each command in `directory`."""

        also = cls._see_also(main_parser)

        # pylint: disable=protected-access
        for action in main_parser._subparsers._actions:
            if isinstance(action, argparse._SubParsersAction):
                for name, parser in action.choices.items():

                    parser.formatter_class = cls.formatter_class
                    helpdoc = parser.format_help()

                    if cls.formatter_class == markdown.MarkdownHelpFormatter:
                        lines = helpdoc.splitlines()
                        see_also = ", ".join([v for k, v in also.items() if k != name]) + "."
                        mdoc = mandown.mandown.Mandown(
                            lines, name=f"lscolors-{name}", see_also=see_also
                        )
                        helpdoc = mdoc.render_markdown() + "\n"

                    pathlib.Path(directory, name + cls.formatter_class_ext).write_text(
                        helpdoc, encoding="utf-8"
                    )

    @staticmethod
    def _see_also(main_parser):
        """Docstring."""

        also = {}

        # pylint: disable=protected-access
        for action in main_parser._subparsers._actions:
            if isinstance(action, argparse._SubParsersAction):
                for name in action.choices:
                    also[name] = f"[lscolors-{name}]({name}.md)"

        return also


# -------------------------------------------------------------------------------


def argformat(parser):
    """Docstring."""
    return Argformat(parser)


configure = Argformat.configure
print_main_page = Argformat.print_main_page
write_command_pages = Argformat.write_command_pages
